import numpy as np

def poisson_2d(f, nx, ny, hx, hy)->np.ndarray:
    """
    Solve a 2D Poisson equation with periodic boundary conditions.
    
    Parameters
    ----------
    f: np.ndarray
        Source term 
    nx: int
        Number of points in the x-direction
    ny: int
        Number of points in the y-direction
    hx: float
        Spatial discretization step in the x-direction
    hy: float
        Spatial discretization step in the y-direction
    
    Returns
    -------
    res: np.ndarray
        The solution of the Poisson problem 
     
    """
    
    # Dimensions (removing duplicate data due to periodicity)
    nnx, nny = nx-1, ny-1
    rny = nny//2+1
    
    # Forward 2D DFT
    F = np.fft.rfftn(f, s=(nnx, nny))
    
    # Compute eigenvalues of Laplace operator in Fourier space with zero mean
    kx = np.fft.fftfreq(nnx, d=hx) * 2 * np.pi
    ky = np.fft.rfftfreq(nny, d=hy) * 2 * np.pi

    den = np.zeros((nnx, rny),dtype=np.float64)
    den[:, :] = kx[:, np.newaxis]**2 + ky[np.newaxis, :]**2
    den[0, 0] = 1.
    F /= -den
    F[0,0] = 0.
    
    # Backward DCT in x and DFT in y, copy data for periodicity
    res = np.empty((nx, ny))
    res[:nnx, :nny] = np.fft.irfftn(F, s=(nnx, nny))
    res[nx-1, :] = res[0, :]
    res[:, ny-1] = res[:, 0]
    return res

def poisson_3d(f : np.ndarray, nx : int, ny : int, nz : int, 
               hx : float, hy : float, hz : float)->np.ndarray:
    """
    Solve a 3D Poisson equation with periodic boundary conditions.
    
    Parameters
    ----------
    f: np.ndarray
        Source term 
    nx: int
        Number of points in the x-direction
    ny: int
        Number of points in the y-direction
    nz: int 
        Number of points in the z-direction
    hx: float
        Spatial discretization step in the x-direction
    hy: float
        Spatial discretization step in the y-direction
    hz: float
        Spatial discretization step in the z-direction
    
    Returns
    -------
    res: np.ndarray
        The solution of the Poisson problem 
    
    """
    
    # Dimensions (removing duplicate data due to periodicity)
    nnx, nny, nnz = nx-1, ny-1, nz-1
    rnz = nnz//2+1
    
    # Forward 3D DFT
    F = np.fft.rfftn(f, s=(nnx, nny, nnz))
    
    # Compute eigenvalues of Laplace operator in Fourier space with zero mean
    kx = np.fft.fftfreq(nnx, d=hx) * 2 * np.pi
    ky = np.fft.fftfreq(nny, d=hy) * 2 * np.pi
    kz = np.fft.fftfreq(nnz, d=hz)[:rnz] * 2 * np.pi

    den = np.empty((nnx,nny,rnz),dtype=np.float64)
    den[:, :, :] = kx[:, np.newaxis, np.newaxis]**2 + ky[np.newaxis, :, np.newaxis]**2 + kz[np.newaxis, np.newaxis, :]**2
    den[0, 0, 0] = 1.
    F /= -den
    F[0,0, 0] = 0.
    
    # Backward DCT in x and DFT in y, copy data for periodicity
    res = np.empty((nx, ny, nz))
    res[:nnx, :nny, :nnz] = np.fft.irfftn(F, s=(nnx, nny, nnz))
    res[nx-1, :, :] = res[0, :, :]
    res[:, ny-1, :] = res[:, 0, :]
    res[:, :, nz-1] = res[:, :, 0]
    return res

def dilatational_velocities(theta : np.ndarray, nx : int, ny : int, nz : int, 
                            hx : float, hy : float, hz : float)->np.ndarray:
    """
    Compute the dilatational velocity field using Fourier space projection.
    
    Parameters
    ----------
    theta: np.ndarray
        Divergence of the velocity field
    nx: int
        Number of points in the x-direction
    ny: int
        Number of points in the y-direction
    nz: int 
        Number of points in the z-direction
    hx: float
        Spatial discretization step in the x-direction
    hy: float
        Spatial discretization step in the y-direction
    hz: float
        Spatial discretization step in the z-direction
    
    Returns
    -------
    vd_x, vd_y, vd_z : np.ndarray
        Dilatational velocity components in real space.
    """
    
    # Dimensions (using full grid size)
    nnx, nny, nnz = nx, ny, nz
    rnz = nnz//2+1
    
    # Forward 3D FFT (using rfftn for real-valued input)
    theta_hat = np.fft.rfftn(theta, s=(nnx, nny, nnz))
    
    # Compute eigenvalues of gradient operator in Fourier space with zero mean
    kx = np.fft.fftfreq(nnx, d=hx) * 2 * np.pi
    ky = np.fft.fftfreq(nny, d=hy) * 2 * np.pi
    kz = np.fft.fftfreq(nnz, d=hz)[:rnz] * 2 * np.pi

    den = np.empty((nnx,nny,rnz),dtype=np.float64)
    den[:, :, :] = kx[:, np.newaxis, np.newaxis]**2 + ky[np.newaxis, :, np.newaxis]**2 + kz[np.newaxis, np.newaxis, :]**2
    den[0, 0, 0] = 1.

    # Compute dilatational velocity components
    vd_x_hat = -1j*theta_hat * kx[:, np.newaxis, np.newaxis] / den
    vd_y_hat = -1j*theta_hat * ky[np.newaxis, :, np.newaxis] / den
    vd_z_hat = -1j*theta_hat * kz[np.newaxis, np.newaxis, :] / den

    # Ensure conjugate symmetry for real-valued input
    vd_x_hat[0, 0, 0] = 0.
    vd_y_hat[0, 0, 0] = 0.
    vd_z_hat[0, 0, 0] = 0.

    # Inverse 3D FFT
    vd_x = np.fft.irfftn(vd_x_hat, s=(nnx, nny, nnz))
    vd_y = np.fft.irfftn(vd_y_hat, s=(nnx, nny, nnz))
    vd_z = np.fft.irfftn(vd_z_hat, s=(nnx, nny, nnz))

    return vd_x, vd_y, vd_z
